# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import sys


from pyface.qt import QtCore, QtGui


from traits.api import DelegatesTo, Instance, on_trait_change, provides


from pyface.tasks.i_advanced_editor_area_pane import IAdvancedEditorAreaPane
from pyface.tasks.i_editor_area_pane import MEditorAreaPane
from .editor_area_pane import EditorAreaDropFilter
from .main_window_layout import MainWindowLayout, PaneItem
from .task_pane import TaskPane
from .util import set_focus

# ----------------------------------------------------------------------------
# 'AdvancedEditorAreaPane' class.
# ----------------------------------------------------------------------------


@provides(IAdvancedEditorAreaPane)
class AdvancedEditorAreaPane(TaskPane, MEditorAreaPane):
    """ The toolkit-specific implementation of an AdvancedEditorAreaPane.

    See the IAdvancedEditorAreaPane interface for API documentation.
    """

    # Private interface ----------------------------------------------------

    _main_window_layout = Instance(MainWindowLayout)

    # ------------------------------------------------------------------------
    # 'TaskPane' interface.
    # ------------------------------------------------------------------------

    def create(self, parent):
        """ Create and set the toolkit-specific control that represents the
            pane.
        """
        self.control = control = EditorAreaWidget(self, parent)
        self._filter = EditorAreaDropFilter(self)
        self.control.installEventFilter(self._filter)

        # Add shortcuts for scrolling through tabs.
        if sys.platform == "darwin":
            next_seq = "Ctrl+}"
            prev_seq = "Ctrl+{"
        else:
            next_seq = "Ctrl+PgDown"
            prev_seq = "Ctrl+PgUp"
        shortcut = QtGui.QShortcut(QtGui.QKeySequence(next_seq), self.control)
        shortcut.activated.connect(self._next_tab)
        shortcut = QtGui.QShortcut(QtGui.QKeySequence(prev_seq), self.control)
        shortcut.activated.connect(self._previous_tab)

        # Add shortcuts for switching to a specific tab.
        mod = "Ctrl+" if sys.platform == "darwin" else "Alt+"
        mapper = QtCore.QSignalMapper(self.control)
        mapper.mapped.connect(self._activate_tab)
        for i in range(1, 10):
            sequence = QtGui.QKeySequence(mod + str(i))
            shortcut = QtGui.QShortcut(sequence, self.control)
            shortcut.activated.connect(mapper.map)
            mapper.setMapping(shortcut, i - 1)

    def destroy(self):
        """ Destroy the toolkit-specific control that represents the pane.
        """
        self.control.removeEventFilter(self._filter)
        self._filter = None

        for editor in self.editors:
            editor_widget = editor.control.parent()
            self.control.destroy_editor_widget(editor_widget)
            editor.editor_area = None

        super(AdvancedEditorAreaPane, self).destroy()

    # ------------------------------------------------------------------------
    # 'IEditorAreaPane' interface.
    # ------------------------------------------------------------------------

    def activate_editor(self, editor):
        """ Activates the specified editor in the pane.
        """
        editor_widget = editor.control.parent()
        editor_widget.setVisible(True)
        editor_widget.raise_()
        editor.control.setFocus()
        self.active_editor = editor

    def add_editor(self, editor):
        """ Adds an editor to the pane.
        """
        editor.editor_area = self
        editor_widget = EditorWidget(editor, self.control)
        self.control.add_editor_widget(editor_widget)
        self.editors.append(editor)

    def remove_editor(self, editor):
        """ Removes an editor from the pane.
        """
        editor_widget = editor.control.parent()
        self.editors.remove(editor)
        self.control.remove_editor_widget(editor_widget)
        editor.editor_area = None
        if not self.editors:
            self.active_editor = None

    # ------------------------------------------------------------------------
    # 'IAdvancedEditorAreaPane' interface.
    # ------------------------------------------------------------------------

    def get_layout(self):
        """ Returns a LayoutItem that reflects the current state of the editors.
        """
        return self._main_window_layout.get_layout_for_area(
            QtCore.Qt.LeftDockWidgetArea
        )

    def set_layout(self, layout):
        """ Applies a LayoutItem to the editors in the pane.
        """
        if layout is not None:
            self._main_window_layout.set_layout_for_area(
                layout, QtCore.Qt.LeftDockWidgetArea
            )

    # ------------------------------------------------------------------------
    # Private interface.
    # ------------------------------------------------------------------------

    def _activate_tab(self, index):
        """ Activates the tab with the specified index, if there is one.
        """
        widgets = self.control.get_dock_widgets_ordered()
        if index < len(widgets):
            self.activate_editor(widgets[index].editor)

    def _next_tab(self):
        """ Activate the tab after the currently active tab.
        """
        if self.active_editor:
            widgets = self.control.get_dock_widgets_ordered()
            index = widgets.index(self.active_editor.control.parent()) + 1
            if index < len(widgets):
                self.activate_editor(widgets[index].editor)

    def _previous_tab(self):
        """ Activate the tab before the currently active tab.
        """
        if self.active_editor:
            widgets = self.control.get_dock_widgets_ordered()
            index = widgets.index(self.active_editor.control.parent()) - 1
            if index >= 0:
                self.activate_editor(widgets[index].editor)

    def _get_label(self, editor):
        """ Return a tab label for an editor.
        """
        label = editor.name
        if editor.dirty:
            label = "*" + label
        return label

    # Trait initializers ---------------------------------------------------

    def __main_window_layout_default(self):
        return EditorAreaMainWindowLayout(editor_area=self)

    # Trait change handlers ------------------------------------------------

    @on_trait_change("editors:[dirty, name]")
    def _update_label(self, editor, name, new):
        editor.control.parent().update_title()

    @on_trait_change("editors:tooltip")
    def _update_tooltip(self, editor, name, new):
        editor.control.parent().update_tooltip()


# ----------------------------------------------------------------------------
# Auxillary classes.
# ----------------------------------------------------------------------------


class EditorAreaMainWindowLayout(MainWindowLayout):
    """ A MainWindowLayout for implementing AdvancedEditorAreaPane.

    Used for getting and setting layouts for the pane.
    """

    # 'MainWindowLayout' interface -----------------------------------------

    control = DelegatesTo("editor_area")

    # 'TaskWindowLayout' interface -----------------------------------------

    editor_area = Instance(AdvancedEditorAreaPane)

    # ------------------------------------------------------------------------
    # 'MainWindowLayout' abstract interface.
    # ------------------------------------------------------------------------

    def _get_dock_widget(self, pane):
        """ Returns the QDockWidget associated with a PaneItem.
        """
        try:
            editor = self.editor_area.editors[pane.id]
            return editor.control.parent()
        except IndexError:
            return None

    def _get_pane(self, dock_widget):
        """ Returns a PaneItem for a QDockWidget.
        """
        for i, editor in enumerate(self.editor_area.editors):
            if editor.control == dock_widget.widget():
                return PaneItem(id=i)
        return None


class EditorAreaWidget(QtGui.QMainWindow):
    """ An auxillary widget for implementing AdvancedEditorAreaPane.
    """

    # ------------------------------------------------------------------------
    # 'EditorAreaWidget' interface.
    # ------------------------------------------------------------------------

    def __init__(self, editor_area, parent=None):
        super(EditorAreaWidget, self).__init__(parent)
        self.editor_area = editor_area
        self.reset_drag()

        # Fish out the rubber band used by Qt to indicate a drop region. We use
        # it to determine which dock widget is the hover widget.
        for child in self.children():
            if isinstance(child, QtGui.QRubberBand):
                child.installEventFilter(self)
                self._rubber_band = child
                break

        # Monitor focus changes so we can set the active editor.
        QtGui.QApplication.instance().focusChanged.connect(self._focus_changed)

        # Configure the QMainWindow.
        # FIXME: Currently animation is not supported.
        self.setAcceptDrops(True)
        self.setAnimated(False)
        self.setDockNestingEnabled(True)
        self.setDocumentMode(True)
        self.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.setTabPosition(
            QtCore.Qt.AllDockWidgetAreas, QtGui.QTabWidget.North
        )

    def add_editor_widget(self, editor_widget):
        """ Adds a dock widget to the editor area.
        """
        editor_widget.installEventFilter(self)
        self.addDockWidget(QtCore.Qt.LeftDockWidgetArea, editor_widget)

        # Try to place the editor in a sensible spot.
        top_left = None
        for widget in self.get_dock_widgets():
            if top_left is None or (
                widget.pos().manhattanLength()
                < top_left.pos().manhattanLength()
            ):
                top_left = widget
        if top_left:
            self.tabifyDockWidget(top_left, editor_widget)
            top_left.set_title_bar(False)

        # Qt will not give the dock widget focus by default.
        self.editor_area.activate_editor(editor_widget.editor)

    def destroy_editor_widget(self, editor_widget):
        """ Destroys a dock widget in the editor area.
        """
        editor_widget.hide()
        editor_widget.removeEventFilter(self)
        editor_widget.editor.destroy()
        self.removeDockWidget(editor_widget)

    def get_dock_widgets(self):
        """ Gets all visible dock widgets.
        """
        return [
            child
            for child in self.children()
            if isinstance(child, QtGui.QDockWidget) and child.isVisible()
        ]

    def get_dock_widgets_for_bar(self, tab_bar):
        """ Get the dock widgets, in order, attached to given tab bar.

        Because QMainWindow locks this info down, we have resorted to a hack.
        """
        pos = tab_bar.pos()
        key = lambda w: QtGui.QVector2D(pos - w.pos()).lengthSquared()
        all_widgets = self.get_dock_widgets()
        if all_widgets:
            current = min(all_widgets, key=key)
            widgets = self.tabifiedDockWidgets(current)
            widgets.insert(tab_bar.currentIndex(), current)
            return widgets
        return []

    def get_dock_widgets_ordered(self, visible_only=False):
        """ Gets all dock widgets in left-to-right, top-to-bottom order.
        """
        children = []
        for child in self.children():
            if (
                child.isWidgetType()
                and child.isVisible()
                and (
                    (isinstance(child, QtGui.QTabBar) and not visible_only)
                    or (
                        isinstance(child, QtGui.QDockWidget)
                        and (
                            visible_only or not self.tabifiedDockWidgets(child)
                        )
                    )
                )
            ):
                children.append(child)
        children.sort(key=lambda _child: (_child.pos().y(), _child.pos().x()))

        widgets = []
        for child in children:
            if isinstance(child, QtGui.QTabBar):
                widgets.extend(self.get_dock_widgets_for_bar(child))
            else:
                widgets.append(child)
        return widgets

    def remove_editor_widget(self, editor_widget):
        """ Removes a dock widget from the editor area.
        """
        # Get the tabs in this editor's dock area before removing it.
        tabified = self.tabifiedDockWidgets(editor_widget)
        if tabified:
            widgets = self.get_dock_widgets_ordered()
            tabified = [
                widget
                for widget in widgets
                if widget in tabified or widget == editor_widget
            ]
        visible = self.get_dock_widgets_ordered(visible_only=True)

        # Destroy and remove the editor. Get the active widget first, since it
        # may be destroyed!
        next_widget = self.editor_area.active_editor.control.parent()
        self.destroy_editor_widget(editor_widget)

        # Ensure that the appropriate editor is activated.
        editor_area = self.editor_area
        choices = tabified if len(tabified) >= 2 else visible
        if len(choices) >= 2 and editor_widget == next_widget:
            i = choices.index(editor_widget)
            next_widget = (
                choices[i + 1] if i + 1 < len(choices) else choices[i - 1]
            )
            editor_area.activate_editor(next_widget.editor)

        # Update tab bar hide state.
        if len(tabified) == 2:
            next_widget.editor.control.parent().set_title_bar(True)
        if editor_area.hide_tab_bar and len(editor_area.editors) == 1:
            editor_area.editors[0].control.parent().set_title_bar(False)

    def reset_drag(self):
        """ Clear out all drag state.
        """
        self._drag_widget = None
        self._hover_widget = None
        self._tear_handled = False
        self._tear_widgets = []

    def set_hover_widget(self, widget):
        """ Set the dock widget being 'hovered over' during a drag.
        """
        old_widget = self._hover_widget
        self._hover_widget = widget

        if old_widget:
            if old_widget in self._tear_widgets:
                if len(self._tear_widgets) == 1:
                    old_widget.set_title_bar(True)
            elif not self.tabifiedDockWidgets(old_widget):
                old_widget.set_title_bar(True)

        if widget:
            if widget in self._tear_widgets:
                if len(self._tear_widgets) == 1:
                    widget.set_title_bar(False)
            elif len(self.tabifiedDockWidgets(widget)) == 1:
                widget.set_title_bar(False)

    # ------------------------------------------------------------------------
    # Event handlers.
    # ------------------------------------------------------------------------

    def childEvent(self, event):
        """ Reimplemented to gain access to the tab bars as they are created.
        """
        super(EditorAreaWidget, self).childEvent(event)
        if event.polished():
            child = event.child()
            if isinstance(child, QtGui.QTabBar):
                # Use UniqueConnections since Qt recycles the tab bars.
                child.installEventFilter(self)
                child.currentChanged.connect(
                    self._tab_index_changed, QtCore.Qt.UniqueConnection
                )
                child.setTabsClosable(True)
                child.setUsesScrollButtons(True)
                child.tabCloseRequested.connect(
                    self._tab_close_requested, QtCore.Qt.UniqueConnection
                )
                # FIXME: We would like to have the tabs movable, but this
                # confuses the QMainWindowLayout. For now, we disable this.
                # child.setMovable(True)

    def eventFilter(self, obj, event):
        """ Reimplemented to dispatch to sub-handlers.
        """
        if isinstance(obj, QtGui.QDockWidget):
            return self._filter_dock_widget(obj, event)

        elif isinstance(obj, QtGui.QRubberBand):
            return self._filter_rubber_band(obj, event)

        elif isinstance(obj, QtGui.QTabBar):
            return self._filter_tab_bar(obj, event)

        return False

    def _filter_dock_widget(self, widget, event):
        """ Support hover widget state tracking.
        """
        if self._drag_widget and event.type() == QtCore.QEvent.Resize:
            if widget.geometry() == self._rubber_band.geometry():
                self.set_hover_widget(widget)

        elif (
            self._drag_widget == widget and event.type() == QtCore.QEvent.Move
        ):
            if len(self._tear_widgets) == 1 and not self._tear_handled:
                widget = self._tear_widgets[0]
                widget.set_title_bar(True)
                self._tear_handled = True

        elif (
            self._drag_widget == widget
            and event.type() == QtCore.QEvent.MouseButtonRelease
        ):
            self.reset_drag()

        return False

    def _filter_rubber_band(self, rubber_band, event):
        """ Support hover widget state tracking.
        """
        if self._drag_widget and event.type() in (
            QtCore.QEvent.Resize,
            QtCore.QEvent.Move,
        ):
            self.set_hover_widget(None)

        return False

    def _filter_tab_bar(self, tab_bar, event):
        """ Support 'tearing off' a tab.
        """
        if event.type() == QtCore.QEvent.MouseMove:
            if tab_bar.rect().contains(event.pos()):
                self.reset_drag()
            else:
                if not self._drag_widget:
                    index = tab_bar.currentIndex()
                    self._tear_widgets = self.get_dock_widgets_for_bar(tab_bar)
                    self._drag_widget = widget = self._tear_widgets.pop(index)

                    pos = QtCore.QPoint(0, 0)
                    press_event = QtGui.QMouseEvent(
                        QtCore.QEvent.MouseButtonPress,
                        pos,
                        widget.mapToGlobal(pos),
                        QtCore.Qt.LeftButton,
                        QtCore.Qt.LeftButton,
                        event.modifiers(),
                    )
                    QtCore.QCoreApplication.sendEvent(widget, press_event)
                    return True

                event = QtGui.QMouseEvent(
                    QtCore.QEvent.MouseMove,
                    event.pos(),
                    event.globalPos(),
                    event.button(),
                    event.buttons(),
                    event.modifiers(),
                )
                QtCore.QCoreApplication.sendEvent(self._drag_widget, event)
                return True

        elif event.type() == QtCore.QEvent.ToolTip:
            # QDockAreaLayout forces the tooltips to be QDockWidget.windowTitle,
            # so we provide the tooltips manually.
            widgets = self.get_dock_widgets_for_bar(tab_bar)
            index = tab_bar.tabAt(event.pos())
            tooltip = widgets[index].editor.tooltip if index >= 0 else ""
            if tooltip:
                QtGui.QToolTip.showText(event.globalPos(), tooltip, tab_bar)
            return True

        return False

    def focusInEvent(self, event):
        """ Assign focus to the active editor, if possible.
        """
        active_editor = self.editor_area.active_editor
        if active_editor:
            set_focus(active_editor.control)

    # ------------------------------------------------------------------------
    # Signal handlers.
    # ------------------------------------------------------------------------

    def _focus_changed(self, old, new):
        """ Handle an application-level focus change.
        """
        if new is not None:
            for editor in self.editor_area.editors:
                control = editor.control
                if control is not None and control.isAncestorOf(new):
                    self.editor_area.active_editor = focused = editor
                    break
            else:
                if not self.editor_area.editors:
                    self.editor_area.active_editor = None

    def _tab_index_changed(self, index):
        """ Handle a tab selection.
        """
        widgets = self.get_dock_widgets_for_bar(self.sender())
        if index < len(widgets):
            editor_widget = widgets[index]
            editor_widget.editor.control.setFocus()

    def _tab_close_requested(self, index):
        """ Handle a tab close request.
        """
        editor_widget = self.get_dock_widgets_for_bar(self.sender())[index]
        editor_widget.editor.close()


class EditorWidget(QtGui.QDockWidget):
    """ An auxillary widget for implementing AdvancedEditorAreaPane.
    """

    def __init__(self, editor, parent=None):
        super(EditorWidget, self).__init__(parent)
        self.editor = editor
        self.editor.create(self)
        self.setAllowedAreas(QtCore.Qt.LeftDockWidgetArea)
        self.setFeatures(
            QtGui.QDockWidget.DockWidgetClosable
            | QtGui.QDockWidget.DockWidgetMovable
        )
        self.setWidget(editor.control)
        self.update_title()

        # Update the minimum size.
        contents_minsize = editor.control.minimumSize()
        style = self.style()
        contents_minsize.setHeight(
            contents_minsize.height()
            + style.pixelMetric(style.PM_DockWidgetHandleExtent)
        )
        self.setMinimumSize(contents_minsize)

        self.dockLocationChanged.connect(self.update_title_bar)
        self.visibilityChanged.connect(self.update_title_bar)

    def update_title(self):
        title = self.editor.editor_area._get_label(self.editor)
        self.setWindowTitle(title)

        title_bar = self.titleBarWidget()
        if isinstance(title_bar, EditorTitleBarWidget):
            title_bar.setTabText(0, title)

    def update_tooltip(self):
        title_bar = self.titleBarWidget()
        if isinstance(title_bar, EditorTitleBarWidget):
            title_bar.setTabToolTip(0, self.editor.tooltip)

    def update_title_bar(self):
        if self not in self.parent()._tear_widgets:
            tabbed = self.parent().tabifiedDockWidgets(self)
            self.set_title_bar(not tabbed)

    def set_title_bar(self, title_bar):
        current = self.titleBarWidget()
        editor_area = self.editor.editor_area
        if (
            title_bar
            and editor_area
            and (not editor_area.hide_tab_bar or len(editor_area.editors) > 1)
        ):
            if not isinstance(current, EditorTitleBarWidget):
                self.setTitleBarWidget(EditorTitleBarWidget(self))
        elif current is None or isinstance(current, EditorTitleBarWidget):
            self.setTitleBarWidget(QtGui.QWidget())


class EditorTitleBarWidget(QtGui.QTabBar):
    """ An auxillary widget for implementing AdvancedEditorAreaPane.
    """

    def __init__(self, editor_widget):
        super(EditorTitleBarWidget, self).__init__(editor_widget)
        self.addTab(editor_widget.windowTitle())
        self.setTabToolTip(0, editor_widget.editor.tooltip)
        self.setDocumentMode(True)
        self.setExpanding(False)
        self.setTabsClosable(True)

        self.tabCloseRequested.connect(editor_widget.editor.close)

    def mousePressEvent(self, event):
        self.parent().parent()._drag_widget = self.parent()
        event.ignore()

    def mouseMoveEvent(self, event):
        event.ignore()

    def mouseReleaseEvent(self, event):
        event.ignore()



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Color utilities. """


from numpy import asarray, array


# fixme: This should move into enable.
def wx_to_enable_color(color):
    """ Convert a wx color spec. to an enable color spec. """

    enable_color = array((1.0, 1.0, 1.0, 1.0))
    enable_color[:3] = asarray(color.Get()) / 255.0

    return tuple(enable_color)



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!


from pyface.action.api import Action, ActionItem, Group
from traits.api import Any, List, Instance, Property, Str, on_trait_change


from pyface.tasks.task import Task
from pyface.tasks.task_window import TaskWindow


class TaskToggleAction(Action):
    """ An action for activating a task.
    """

    # 'Action' interface ---------------------------------------------------

    #: The user-visible name of the action, matches the task name.
    name = Property(Str, depends_on="task.name")

    #: The action is a toggle menu item.
    style = "toggle"

    #: The tooltip to display for the menu item.
    tooltip = Property(Str, depends_on="name")

    # 'TaskActivateAction' interface ---------------------------------------

    task = Instance(Task)

    # ------------------------------------------------------------------------
    # 'Action' interface.
    # ------------------------------------------------------------------------

    def destroy(self):
        super(TaskToggleAction, self).destroy()

        # Make sure that we are not listening to changes in the task anymore
        # In traits style, we will set the basic object to None and have the
        # listener check that if it is still there.
        self.task = None

    def perform(self, event=None):
        window = self.task.window
        window.activate_task(self.task)

    # ------------------------------------------------------------------------
    # Private interface.
    # ------------------------------------------------------------------------

    def _get_name(self):
        if self.task is None:
            return "UNDEFINED"
        return self.task.name

    def _get_tooltip(self):
        return u"Switch to the %s task." % self.name

    @on_trait_change("task.window.active_task")
    def _update_checked(self):
        if self.task:
            window = self.task.window
            self.checked = (
                window is not None and window.active_task == self.task
            )


class TaskToggleGroup(Group):
    """ A menu for changing the active task in a task window.
    """

    # 'ActionManager' interface --------------------------------------------

    id = "TaskToggleGroup"
    items = List()

    # 'TaskChangeMenuManager' interface ------------------------------------

    # The ActionManager to which the group belongs.
    manager = Any()

    # The window that contains the group.
    window = Instance(TaskWindow)

    # ------------------------------------------------------------------------
    # Private interface.
    # ------------------------------------------------------------------------

    def _get_items(self):
        items = []
        if len(self.window.tasks) > 1:
            # at least two tasks, so something to toggle
            items = [
                ActionItem(action=TaskToggleAction(task=task))
                for task in self.window.tasks
            ]
        return items

    def _rebuild(self):
        # Clear out the old group, then build the new one.
        self.destroy()
        self.items = self._get_items()

        # Inform our manager that it needs to be rebuilt.
        self.manager.changed = True

    # Trait initializers ---------------------------------------------------

    def _items_default(self):
        self.window.on_trait_change(self._rebuild, "tasks[]")
        return self._get_items()

    def _manager_default(self):
        manager = self
        while isinstance(manager, Group):
            manager = manager.parent
        return manager

    def _window_default(self):
        return self.manager.controller.task.window



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import wx

from .default_renderer import DefaultRenderer


class FontRenderer(DefaultRenderer):
    """Render data in the specified color and font and fontsize.
    """

    def DrawForeground(self, grid, attr, dc, rect, row, col, isSelected):
        text = grid.model.GetValue(row, col)
        dc.SetTextForeground(self.color)
        dc.SetFont(self.font)
        dc.DrawText(text, rect.x + 1, rect.y + 1)

    def DrawOld(self, grid, attr, dc, rect, row, col, isSelected):
        # Here we draw text in a grid cell using various fonts
        # and colors.  We have to set the clipping region on
        # the grid's DC, otherwise the text will spill over
        # to the next cell
        dc.SetClippingRegion(rect)

        # clear the background
        dc.SetBackgroundMode(wx.SOLID)
        if isSelected:
            dc.SetBrush(wx.Brush(wx.BLUE, wx.SOLID))
            dc.SetPen(wx.Pen(wx.BLUE, 1, wx.SOLID))
        else:
            dc.SetBrush(wx.Brush(wx.WHITE, wx.SOLID))
            dc.SetPen(wx.Pen(wx.WHITE, 1, wx.SOLID))
        dc.DrawRectangle(rect.x, rect.y, rect.width, rect.height)

        text = grid.model.GetValue(row, col)
        dc.SetBackgroundMode(wx.SOLID)

        # change the text background based on whether the grid is selected
        # or not
        if isSelected:
            dc.SetBrush(self.selectedBrush)
            dc.SetTextBackground("blue")
        else:
            dc.SetBrush(self.normalBrush)
            dc.SetTextBackground("white")

        dc.SetTextForeground(self.color)
        dc.SetFont(self.font)
        dc.DrawText(text, rect.x + 1, rect.y + 1)

        # Okay, now for the advanced class :)
        # Let's add three dots "..."
        # to indicate that that there is more text to be read
        # when the text is larger than the grid cell

        width, height = dc.GetTextExtent(text)
        if width > rect.width - 2:
            width, height = dc.GetTextExtent("...")
            x = rect.x + 1 + rect.width - 2 - width
            dc.DrawRectangle(x, rect.y + 1, width + 1, height)
            dc.DrawText("...", x, rect.y + 1)

        dc.DestroyClippingRegion()


class FontRendererFactory88(object):
    """ I don't grok why this Factory (which I copied from the wx demo)
        was ever necessary? """

    def __init__(self, color, font, fontsize):
        """
        (color, font, fontsize) -> set of a factory to generate
        renderers when called.
        func = MegaFontRenderFactory(color, font, fontsize)
        renderer = func(table)
        """
        self.color = color
        self.font = font
        self.fontsize = fontsize

    def __call__(self, table):
        return FontRenderer(table, self.color, self.font, self.fontsize)



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!


from pyface.qt import QtCore, QtGui

from traits.api import Any, List, Str, provides

from pyface.constant import CANCEL
from pyface.i_single_choice_dialog import (
    ISingleChoiceDialog,
    MSingleChoiceDialog,
)
from .dialog import Dialog, _RESULT_MAP


@provides(ISingleChoiceDialog)
class SingleChoiceDialog(MSingleChoiceDialog, Dialog):
    """ A dialog that allows the user to chose a single item from a list.

    Note that due to limitations of the QInputDialog class, the cancel trait
    is ignored, and the list of displayed strings must be unique.
    """

    # 'ISingleChoiceDialog' interface -------------------------------------#

    #: List of objects to choose from.
    choices = List(Any)

    #: The object chosen, if any.
    choice = Any()

    #: An optional attribute to use for the name of each object in the dialog.
    name_attribute = Str()

    #: The message to display to the user.
    message = Str()

    def set_dialog_choice(self, choice):
        if self.control is not None:
            if self.name_attribute != "":
                choice = getattr(choice, self.name_attribute)
            self.control.setTextValue(str(choice))

    # ------------------------------------------------------------------------
    # Protected 'IDialog' interface.
    # ------------------------------------------------------------------------

    def _create_contents(self, parent):
        """ Creates the window contents. """
        # In this case, Qt does it all for us in 'QInputDialog'
        pass

    def _show_modal(self):
        self.control.setWindowModality(QtCore.Qt.ApplicationModal)
        retval = self.control.exec_()
        if self.control is None:
            # dialog window closed, treat as Cancel, nullify choice
            retval = CANCEL
        else:
            retval = _RESULT_MAP[retval]
        return retval

    # ------------------------------------------------------------------------
    # 'IWindow' interface.
    # ------------------------------------------------------------------------

    def close(self):
        """ Closes the window. """

        # Get the chosen object.
        if self.control is not None and self.return_code != CANCEL:
            text = self.control.textValue()
            choices = self._choice_strings()
            if text in choices:
                idx = self._choice_strings().index(text)
                self.choice = self.choices[idx]
            else:
                self.choice = None

        # Let the window close as normal.
        super(SingleChoiceDialog, self).close()

    # ------------------------------------------------------------------------
    # Protected 'IWidget' interface.
    # ------------------------------------------------------------------------

    def _create_control(self, parent):
        """ Create the toolkit-specific control that represents the window. """
        dialog = QtGui.QInputDialog(parent)

        dialog.setOption(QtGui.QInputDialog.UseListViewForComboBoxItems, True)
        dialog.setWindowTitle(self.title)
        dialog.setLabelText(self.message)

        # initialize choices: set initial value to first choice
        choices = self._choice_strings()
        dialog.setComboBoxItems(choices)
        dialog.setTextValue(choices[0])

        if self.size != (-1, -1):
            self.resize(*self.size)
        if self.position != (-1, -1):
            self.move(*self.position)

        return dialog



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Test the scripting tools. """


import unittest
from pyface.util.id_helper import get_unique_id, object_counter


class IDHelperTestCase(unittest.TestCase):
    """ Test the scripting tools. """

    # Tests ----------------------------------------------------------------

    def test_object_counter(self):

        from traits.api import WeakRef

        class Bogus(object):
            weak = WeakRef

        class Foo(object):
            foo = 3

        foo = Foo()

        self.assertEqual(object_counter.get_count(Bogus), 0)
        self.assertEqual(object_counter.next_count(Bogus), 1)
        self.assertEqual(object_counter.next_count(Bogus), 2)
        self.assertEqual(object_counter.get_count(Bogus), 2)
        self.assertEqual(object_counter.next_count(foo), 1)
        self.assertEqual(object_counter.next_count(Bogus), 3)

    def test_get_unique_id(self):
        class Bogus(object):
            pass

        bogus_1 = Bogus()
        bogus_2 = Bogus()

        self.assertEqual(get_unique_id(bogus_1), "Bogus_1")
        self.assertEqual(get_unique_id(bogus_2), "Bogus_2")


if __name__ == "__main__":
    unittest.main()



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!


from .dock_pane_toggle_group import DockPaneToggleGroup
from .schema import (
    ActionSchema,
    GroupSchema,
    MenuSchema,
    MenuBarSchema,
    ToolBarSchema,
    SGroup,
    SMenu,
    SMenuBar,
    SToolBar,
)
from .schema_addition import SchemaAddition
from .task_action import (
    CentralPaneAction,
    DockPaneAction,
    EditorAction,
    TaskAction,
    TaskWindowAction,
)
from .task_action_controller import TaskActionController
from .task_action_manager_builder import TaskActionManagerBuilder
from .task_toggle_group import TaskToggleGroup
from .task_window_toggle_group import TaskWindowToggleGroup
from .tasks_application_action import (
    CreateTaskWindowAction,
    TasksApplicationAction,
)



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Pyface 'DockWindow' support.

    This package provides a Pyface 'dockable' window component that allows
    child windows to be reorganized within the DockWindow using drag and drop.
    The component also allows multiple sub-windows to occupy the same
    sub-region of the DockWindow, in which case each sub-window appears as a
    separate notebook-like tab within the region.
"""



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Abstract base class for a node in a preference dialog.

A preference node has a label and an image which are used to represent the node
in a preference dialog (usually in the form of a tree).

"""


from pyface.viewer.tree_item import TreeItem
from traits.api import Str


class PreferenceNode(TreeItem):
    """ Abstract base class for a node in a preference dialog.

    A preference node has a name and an image which are used to represent the
    node in a preference dialog (usually in the form of a tree).

    """

    # 'PreferenceNode' interface -------------------------------------------

    # The node's unique Id.
    id = Str()

    # The node's image.
    image = Str()

    # The node name.
    name = Str()

    # The Id of the help topic for the node.
    help_id = Str()

    # ------------------------------------------------------------------------
    # 'object' interface.
    # ------------------------------------------------------------------------

    def __str__(self):
        """ Returns the string representation of the item. """

        return self.name

    # ------------------------------------------------------------------------
    # 'PreferenceNode' interface.
    # ------------------------------------------------------------------------

    def create_page(self):
        """ Creates the preference page for this node. """

        raise NotImplementedError

    def lookup(self, id):
        """ Returns the child of this node with the specified Id.

        Returns None if no such child exists.

        """

        for node in self.children:
            if node.id == id:
                break

        else:
            node = None

        return node



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" The Wx-specific implementation of the text field class """


import wx

from traits.api import provides

from pyface.fields.i_text_field import ITextField, MTextField
from .field import Field


@provides(ITextField)
class TextField(MTextField, Field):
    """ The Wx-specific implementation of the text field class """

    # ------------------------------------------------------------------------
    # IWidget interface
    # ------------------------------------------------------------------------

    def _create_control(self, parent):
        """ Create the toolkit-specific control that represents the widget. """

        style = wx.TE_PROCESS_ENTER
        if self.echo == "password":
            style |= wx.TE_PASSWORD

        control = wx.TextCtrl(parent, -1, value=self.value, style=style)
        return control

    # ------------------------------------------------------------------------
    # Private interface
    # ------------------------------------------------------------------------

    def _update_value(self, event):
        # do normal focus event stuff
        if isinstance(event, wx.FocusEvent):
            event.Skip()
        if self.control is not None:
            self.value = self.control.GetValue()

    def _get_control_value(self):
        """ Toolkit specific method to get the control's value. """
        return self.control.GetValue()

    def _set_control_value(self, value):
        """ Toolkit specific method to set the control's value. """
        self.control.SetValue(value)

    def _observe_control_value(self, remove=False):
        """ Toolkit specific method to change the control value observer. """
        if remove:
            self.control.Unbind(wx.EVT_TEXT, handler=self._update_value)
        else:
            self.control.Bind(wx.EVT_TEXT, self._update_value)

    def _get_control_placeholder(self):
        """ Toolkit specific method to set the control's placeholder. """
        return self.control.GetHint()

    def _set_control_placeholder(self, placeholder):
        """ Toolkit specific method to set the control's placeholder. """
        self.control.SetHint(placeholder)

    def _get_control_echo(self):
        """ Toolkit specific method to get the control's echo. """
        return self.echo

    def _set_control_echo(self, echo):
        """ Toolkit specific method to set the control's echo. """
        # Can't change echo on Wx after control has been created."
        pass

    def _get_control_read_only(self):
        """ Toolkit specific method to get the control's read_only state. """
        return not self.control.IsEditable()

    def _set_control_read_only(self, read_only):
        """ Toolkit specific method to set the control's read_only state. """
        self.control.SetEditable(not read_only)

    def _observe_control_editing_finished(self, remove=False):
        """ Change observation of whether editing is finished. """
        if remove:
            self.control.Unbind(wx.EVT_TEXT_ENTER, handler=self._update_value)
        else:
            self.control.Bind(wx.EVT_TEXT_ENTER, self._update_value)



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" A tree control with a model/ui architecture. """


import logging
import os


import wx


from traits.api import (
    Any,
    Bool,
    Callable,
    Enum,
    Event,
    Instance,
    List,
    Property,
    Str,
)


from pyface.filter import Filter
from pyface.key_pressed_event import KeyPressedEvent
from pyface.sorter import Sorter
from pyface.tree.tree_model import TreeModel
from pyface.ui.wx.gui import GUI
from pyface.ui.wx.image_list import ImageList
from pyface.ui.wx.widget import Widget
from pyface.wx.drag_and_drop import PythonDropSource, PythonDropTarget


# Create a logger for this module.
logger = logging.getLogger(__name__)


class _Tree(wx.TreeCtrl):
    """ The wx tree control that we delegate to.

    We use this derived class so that we can detect the destruction of the
    tree and remove model listeners etc.

    """

    # ------------------------------------------------------------------------
    # 'object' interface.
    # ------------------------------------------------------------------------

    def __init__(self, tree, parent, wxid, style):
        """ Creates a new tree. """

        # Base class constructor.
        super().__init__(parent, wxid, style=style)

        # The tree that we are the toolkit-specific delegate for.
        self._tree = tree

    def Destroy(self):
        """ Destructor. """

        # Stop listenting to the model!
        self._tree._remove_model_listeners(self._tree.model)

        super().Destroy()


class Tree(Widget):
    """ A tree control with a model/ui architecture. """

    # The default tree style.
    STYLE = wx.TR_EDIT_LABELS | wx.TR_HAS_BUTTONS | wx.CLIP_CHILDREN

    # 'Tree' interface -----------------------------------------------------

    # The tree's filters (empty if no filtering is required).
    filters = List(Instance(Filter))

    # Mode for lines connecting tree nodes which emphasize hierarchy:
    # 'appearance' - only on when lines look good,
    # 'on' - always on, 'off' - always off
    # NOTE: on and off are ignored in favor of show_lines for now
    lines_mode = Enum("appearance", "on", "off")

    # The model that provides the data for the tree.
    model = Instance(TreeModel, ())

    # The root of the tree (this is for convenience, it just delegates to
    # the tree's model).
    root = Property(Any)

    # The objects currently selected in the tree.
    selection = List(Any)

    # Selection mode.
    selection_mode = Enum("single", "extended")

    # Should an image be shown for each node?
    show_images = Bool(True)

    # Should lines be drawn between levels in the tree.
    show_lines = Bool(True)

    # Should the root of the tree be shown?
    show_root = Bool(True)

    # The tree's sorter (None if no sorting is required).
    sorter = Instance(Sorter)

    # Events ----

    # A right-click occurred on the control (not a node!).
    control_right_clicked = Event  # (Point)

    # A key was pressed while the tree has focus.
    key_pressed = Event(Instance(KeyPressedEvent))

    # A node has been activated (ie. double-clicked).
    node_activated = Event  # (Any)

    # A drag operation was started on a node.
    node_begin_drag = Event  # (Any)

    # A (non-leaf) node has been collapsed.
    node_collapsed = Event  # (Any)

    # A (non-leaf) node has been expanded.
    node_expanded = Event  # (Any)

    # A left-click occurred on a node.
    #
    # Tuple(node, point).
    node_left_clicked = Event  # (Tuple)

    # A right-click occurred on a node.
    #
    # Tuple(node, point)
    node_right_clicked = Event  # (Tuple)

    # Private interface ----------------------------------------------------

    # A name to distinguish the tree for debugging!
    #
    # fixme: This turns out to be kinda useful... Should 'Widget' have a name
    # trait?
    _name = Str("Anonymous tree")

    # An optional callback to detect the end of a label edit.  This is
    # useful because the callback will be invoked even if the node label was
    # not actually changed.
    _label_edit_callback = Callable

    # Flag for allowing selection events to be ignored
    _ignore_selection_events = Bool(False)

    # ------------------------------------------------------------------------
    # 'object' interface.
    # ------------------------------------------------------------------------

    def __init__(self, parent, image_size=(16, 16), **traits):
        """ Creates a new tree.

        'parent' is the toolkit-specific control that is the tree's parent.

        'image_size' is a tuple in the form (int width, int height) that
        specifies the size of the images (if required) displayed in the tree.

        """

        # Base class constructors.
        super(Tree, self).__init__(**traits)

        # Create the toolkit-specific control.
        self.control = tree = _Tree(
            self, parent, style=self._get_style()
        )

        # Wire up the wx tree events.
        tree.Bind(wx.EVT_CHAR, self._on_char)
        tree.Bind(wx.EVT_LEFT_DOWN, self._on_left_down)
        # fixme: This is not technically correct as context menus etc should
        # appear on a right up (or right click).  Unfortunately,  if we
        # change this to 'EVT_RIGHT_UP' wx does not fire the event unless the
        # right mouse button is double clicked 8^()  Sad,  but true!
        tree.Bind(wx.EVT_RIGHT_DOWN, self._on_right_down)
        # fixme: This is not technically correct as we would really like to use
        # 'EVT_TREE_ITEM_ACTIVATED'. Unfortunately, (in 2.6 at least), it
        # throws an exception when the 'Enter' key is pressed as the wx tree
        # item Id in the event seems to be invalid. It also seems to cause
        # any child frames that my be created in response to the event to
        # appear *behind* the parent window, which is, errrr, not great ;^)
        tree.Bind(wx.EVT_LEFT_DCLICK, self._on_tree_item_activated)
        tree.Bind(wx.EVT_TREE_ITEM_COLLAPSING, self._on_tree_item_collapsing)
        tree.Bind(wx.EVT_TREE_ITEM_COLLAPSED, self._on_tree_item_collapsed)
        tree.Bind(wx.EVT_TREE_ITEM_EXPANDING, self._on_tree_item_expanding)
        tree.Bind(wx.EVT_TREE_ITEM_EXPANDED, self._on_tree_item_expanded)
        tree.Bind(wx.EVT_TREE_BEGIN_LABEL_EDIT, self._on_tree_begin_label_edit)
        tree.Bind(wx.EVT_TREE_END_LABEL_EDIT, self._on_tree_end_label_edit)
        tree.Bind(wx.EVT_TREE_BEGIN_DRAG, self._on_tree_begin_drag)
        tree.Bind(wx.EVT_TREE_SEL_CHANGED, self._on_tree_sel_changed)
        tree.Bind(wx.EVT_TREE_DELETE_ITEM, self._on_tree_delete_item)

        # Enable the tree as a drag and drop target.
        self.control.SetDropTarget(PythonDropTarget(self))

        # The image list is a wxPython-ism that caches all images used in the
        # control.
        self._image_list = ImageList(image_size[0], image_size[1])
        if self.show_images:
            tree.AssignImageList(self._image_list)

        # Mapping from node to wx tree item Ids.
        self._node_to_id_map = {}

        # Add the root node.
        if self.root is not None:
            self._add_root_node(self.root)

        # Listen for changes to the model.
        self._add_model_listeners(self.model)

    # ------------------------------------------------------------------------
    # 'Tree' interface.
    # ------------------------------------------------------------------------

    # Properties -----------------------------------------------------------

    def _get_root(self):
        """ Returns the root node of the tree. """

        return self.model.root

    def _set_root(self, root):
        """ Sets the root node of the tree. """

        self.model.root = root

        return

    # Methods -------------------------------------------------------------#

    def collapse(self, node):
        """ Collapses the specified node. """

        wxid = self._get_wxid(node)
        if wxid is not None:
            self.control.Collapse(wxid)

    def edit_label(self, node, callback=None):
        """ Edits the label of the specified node.

        If a callback is specified it will be called when the label edit
        completes WHETHER OR NOT the label was actually changed.

        The callback must take exactly 3 arguments:- (tree, node, label)

        """

        wxid = self._get_wxid(node)
        if wxid is not None:
            self._label_edit_callback = callback
            self.control.EditLabel(wxid)

    def expand(self, node):
        """ Expands the specified node. """

        wxid = self._get_wxid(node)
        if wxid is not None:
            self.control.Expand(wxid)

    def expand_all(self):
        """ Expands every node in the tree. """

        if self.show_root:
            self._expand_item(self._get_wxid(self.root))

        else:
            for child in self._get_children(self.root):
                self._expand_item(self._get_wxid(child))

    def get_parent(self, node):
        """ Returns the parent of a node.

        This will only work iff the node has been displayed in the tree.  If it
        hasn't then None is returned.

        """

        # Has the node actually appeared in the tree yet?
        wxid = self._get_wxid(node)
        if wxid is not None:
            pid = self.control.GetItemParent(wxid)

            # The item data is a tuple.  The first element indicates whether or
            # not we have already populated the item with its children.  The
            # second element is the actual item data.
            populated, parent = self.control.GetItemData(pid)

        else:
            parent = None

        return parent

    def is_expanded(self, node):
        """ Returns True if the node is expanded, otherwise False. """

        wxid = self._get_wxid(node)
        if wxid is not None:
            # If the root node is hidden then it is always expanded!
            if node is self.root and not self.show_root:
                is_expanded = True

            else:
                is_expanded = self.control.IsExpanded(wxid)

        else:
            is_expanded = False

        return is_expanded

    def is_selected(self, node):
        """ Returns True if the node is selected, otherwise False. """

        wxid = self._get_wxid(node)
        if wxid is not None:
            is_selected = self.control.IsSelected(wxid)

        else:
            is_selected = False

        return is_selected

    def refresh(self, node):
        """ Refresh the tree starting from the specified node.

        Call this when the structure of the content has changed DRAMATICALLY.

        """

        # Has the node actually appeared in the tree yet?
        pid = self._get_wxid(node)
        if pid is not None:
            # Delete all of the node's children and re-add them.
            self.control.DeleteChildren(pid)
            self.control.SetItemData(pid, (False, node))

            # Does the node have any children?
            has_children = self._has_children(node)
            self.control.SetItemHasChildren(pid, has_children)

            # fixme: At least on Windows, wx does not fire an expanding
            # event for a hidden root node, so we have to populate the node
            # manually.
            if node is self.root and not self.show_root:
                # Add the child nodes.
                for child in self._get_children(node):
                    self._add_node(pid, child)

            else:
                # Expand it.
                if self.control.IsExpanded(pid):
                    self.control.Collapse(pid)

                self.control.Expand(pid)

    def select(self, node):
        """ Selects the specified node. """

        wxid = self._get_wxid(node)
        if wxid is not None:
            self.control.SelectItem(wxid)

    def set_selection(self, list):
        """ Selects the specified list of nodes. """
        logger.debug("Setting selection to [%s] within Tree [%s]", list, self)

        # Update the control to reflect the target list by unselecting
        # everything and then selecting each item in the list.  During this
        # process, we want to avoid changing our own selection.
        self._ignore_selection_events = True
        self.control.UnselectAll()
        for node in list:
            try:
                self.select(node)
            except:
                logger.exception("Unable to select node [%s]", node)

        self._ignore_selection_events = False

        # Update our selection to reflect the final selection state.
        self.selection = self._get_selection()

    # ------------------------------------------------------------------------
    # 'PythonDropTarget' interface.
    # ------------------------------------------------------------------------

    def on_drag_over(self, x, y, obj, default_drag_result):
        """ Called when a node is dragged over the tree. """

        result = wx.DragNone

        # Find the node that we are dragging over...
        node = self._get_drag_drop_node(x, y)
        if node is not None:
            # Ask the model if the node allows the object to be dropped onto
            # it.
            if self.model.can_drop(node, obj):
                result = default_drag_result

        return result

    def on_drop(self, x, y, obj, default_drag_result):
        """ Called when a node is dropped on the tree. """

        # Find the node that we are dragging over...
        node = self._get_drag_drop_node(x, y)
        if node is not None:
            self.model.drop(node, obj)

        return default_drag_result

    # ------------------------------------------------------------------------
    # Private interface.
    # ------------------------------------------------------------------------

    def _get_wxid(self, node):
        """ Returns the wxid for the specified node.

        Returns None if the node has not yet appeared in the tree.

        """

        # The model must generate a unique key for each node (unique within the
        # model).
        key = self.model.get_key(node)

        return self._node_to_id_map.get(key, None)

    def _set_wxid(self, node, wxid):
        """ Sets the wxid for the specified node. """

        # The model must generate a unique key for each node (unique within the
        # model).
        key = self.model.get_key(node)

        self._node_to_id_map[key] = wxid

    def _remove_wxid(self, node):
        """ Removes the wxid for the specified node. """

        # The model must generate a unique key for each node (unique within the
        # model).
        key = self.model.get_key(node)

        try:
            del self._node_to_id_map[key]

        except KeyError:
            # fixme: No, really, this is a serious one... How do we get in this
            # situation.  It came up when using the canvas stuff...
            logger.warn("removing node: %s" % str(node))

    def _get_style(self):
        """ Returns the wx style flags for creating the tree control. """

        # Start with the default flags.
        style = self.STYLE

        # Turn lines off for appearance on *nix.
        # ...for now, show_lines determines if lines are on or off, but
        # eventually lines_mode may eliminate the need for show_lines
        if self.lines_mode == "appearance" and os.name == "posix":
            self.show_lines = False

        if not self.show_lines:
            style = style | wx.TR_NO_LINES

        if not self.show_root:
            # fixme: It looks a little weird, but it we don't have the
            # 'lines at root' style then wx won't draw the expand/collapse
            # image on non-leaf nodes at the root level 8^()
            style = style | wx.TR_HIDE_ROOT | wx.TR_LINES_AT_ROOT

        if self.selection_mode != "single":
            style = style | wx.TR_MULTIPLE | wx.TR_EXTENDED

        return style

    def _add_model_listeners(self, model):
        """ Adds listeners for model changes. """

        # Listen for changes to the model.
        model.on_trait_change(self._on_root_changed, "root")
        model.on_trait_change(self._on_nodes_changed, "nodes_changed")
        model.on_trait_change(self._on_nodes_inserted, "nodes_inserted")
        model.on_trait_change(self._on_nodes_removed, "nodes_removed")
        model.on_trait_change(self._on_nodes_replaced, "nodes_replaced")
        model.on_trait_change(self._on_structure_changed, "structure_changed")

    def _remove_model_listeners(self, model):
        """ Removes listeners for model changes. """

        # Unhook the model event listeners.
        model.on_trait_change(self._on_root_changed, "root", remove=True)

        model.on_trait_change(
            self._on_nodes_changed, "nodes_changed", remove=True
        )

        model.on_trait_change(
            self._on_nodes_inserted, "nodes_inserted", remove=True
        )

        model.on_trait_change(
            self._on_nodes_removed, "nodes_removed", remove=True
        )

        model.on_trait_change(
            self._on_nodes_replaced, "nodes_replaced", remove=True
        )

        model.on_trait_change(
            self._on_structure_changed, "structure_changed", remove=True
        )

        # unwire the wx tree events.
        tree = self.control
        tree.Unbind(wx.EVT_CHAR)
        tree.Unbind(wx.EVT_LEFT_DOWN)
        # fixme: This is not technically correct as context menus etc should
        # appear on a right up (or right click).  Unfortunately,  if we
        # change this to 'EVT_RIGHT_UP' wx does not fire the event unless the
        # right mouse button is double clicked 8^()  Sad,  but true!
        tree.Unbind(wx.EVT_RIGHT_DOWN)
        # fixme: This is not technically correct as we would really like to use
        # 'EVT_TREE_ITEM_ACTIVATED'. Unfortunately, (in 2.6 at least), it
        # throws an exception when the 'Enter' key is pressed as the wx tree
        # item Id in the event seems to be invalid. It also seems to cause
        # any child frames that my be created in response to the event to
        # appear *behind* the parent window, which is, errrr, not great ;^)
        tree.Unbind(wx.EVT_LEFT_DCLICK)
        tree.Unbind(wx.EVT_TREE_ITEM_ACTIVATED)
        tree.Unbind(wx.EVT_TREE_ITEM_COLLAPSING)
        tree.Unbind(wx.EVT_TREE_ITEM_COLLAPSED)
        tree.Unbind(wx.EVT_TREE_ITEM_EXPANDING)
        tree.Unbind(wx.EVT_TREE_ITEM_EXPANDED)
        tree.Unbind(wx.EVT_TREE_BEGIN_LABEL_EDIT)
        tree.Unbind(wx.EVT_TREE_END_LABEL_EDIT)
        tree.Unbind(wx.EVT_TREE_BEGIN_DRAG)
        tree.Unbind(wx.EVT_TREE_SEL_CHANGED)
        tree.Unbind(wx.EVT_TREE_DELETE_ITEM)

    def _add_root_node(self, node):
        """ Adds the root node. """

        # Get the tree item image index and the label text.
        image_index = self._get_image_index(node)
        text = self._get_text(node)

        # Add the node.
        wxid = self.control.AddRoot(text, image_index, image_index)

        # This gives the model a chance to wire up trait handlers etc.
        self.model.add_listener(node)

        # If the root node is hidden, get its children.
        if not self.show_root:
            # Add the child nodes.
            for child in self._get_children(node):
                self._add_node(wxid, child)

        # Does the node have any children?
        has_children = self._has_children(node)
        self.control.SetItemHasChildren(wxid, has_children)

        # The item data is a tuple.  The first element indicates whether or not
        # we have already populated the item with its children.  The second
        # element is the actual item data (which in our case is an arbitrary
        # Python object provided by the tree model).
        if self.show_root:
            self.control.SetItemData(wxid, (not self.show_root, node))

        # Make sure that we can find the node's Id.
        self._set_wxid(node, wxid)

        # Automatically expand the root.
        if self.show_root:
            self.control.Expand(wxid)

    def _add_node(self, pid, node):
        """ Adds 'node' as a child of the node identified by 'pid'.

        If 'pid' is None then we are adding the root node.

        """

        # Get the tree item image index and the label text.
        image_index = self._get_image_index(node)
        text = self._get_text(node)

        # Add the node.
        wxid = self.control.AppendItem(pid, text, image_index, image_index)

        # This gives the model a chance to wire up trait handlers etc.
        self.model.add_listener(node)

        # Does the node have any children?
        has_children = self._has_children(node)
        self.control.SetItemHasChildren(wxid, has_children)

        # The item data is a tuple.  The first element indicates whether or not
        # we have already populated the item with its children.  The second
        # element is the actual item data (which in our case is an arbitrary
        # Python object provided by the tree model).
        self.control.SetItemData(wxid, (False, node))

        # Make sure that we can find the node's Id.
        self._set_wxid(node, wxid)

    def _insert_node(self, pid, node, index):
        """ Inserts 'node' as a child of the node identified by 'pid'.

        If 'pid' is None then we are adding the root node.

        """

        # Get the tree item image index and the label text.
        image_index = self._get_image_index(node)
        text = self._get_text(node)

        # Add the node.
        wxid = self.control.Sizer.InsertBefore(
            pid, index, text, image_index, image_index
        )

        # This gives the model a chance to wire up trait handlers etc.
        self.model.add_listener(node)

        # Does the node have any children?
        has_children = self._has_children(node)
        self.control.SetItemHasChildren(wxid, has_children)

        # The item data is a tuple.  The first element indicates whether or not
        # we have already populated the item with its children.  The second
        # element is the actual item data (which in our case is an arbitrary
        # Python object provided by the tree model).
        self.control.SetItemData(wxid, (False, node))

        # Make sure that we can find the node's Id.
        self._set_wxid(node, wxid)

    def _remove_node(self, wxid, node):
        """ Removes a node from the tree. """

        # This gives the model a chance to remove trait handlers etc.
        self.model.remove_listener(node)

        # Remove the reference to the item's data.
        self._remove_wxid(node)
        self.control.SetItemData(wxid, None)

    def _update_node(self, wxid, node):
        """ Updates the image and text of the specified node. """

        # Get the tree item image index.
        image_index = self._get_image_index(node)
        self.control.SetItemImage(wxid, image_index, wx.TreeItemIcon_Normal)
        self.control.SetItemImage(wxid, image_index, wx.TreeItemIcon_Selected)

        # Get the tree item text.
        text = self._get_text(node)
        self.control.SetItemText(wxid, text)

    def _has_children(self, node):
        """ Returns True if a node has children. """

        # fixme: To be correct we *should* apply filtering here, but that
        # seems to blow a hole throught models that have some efficient
        # mechanism for determining whether or not they have children.  There
        # is also a precedent for doing it this way in Windoze, where a node
        # gets marked as though it can be expanded, even thought when the
        # expansion occurs, no children are present!
        return self.model.has_children(node)

    def _get_children(self, node):
        """ Get the children of a node. """

        children = self.model.get_children(node)

        # Filtering....
        filtered_children = []
        for child in children:
            for filter in self.filters:
                if not filter.select(self, node, child):
                    break

            else:
                filtered_children.append(child)

        # Sorting...
        if self.sorter is not None:
            self.sorter.sort(self, node, filtered_children)

        return filtered_children

    def _get_image_index(self, node):
        """ Returns the tree item image index for a node. """

        expanded = self.is_expanded(node)
        selected = self.is_selected(node)

        # Get the image used to represent the node.
        image = self.model.get_image(node, selected, expanded)
        if image is not None:
            image_index = self._image_list.GetIndex(image)

        else:
            image_index = -1

        return image_index

    def _get_drag_drop_node(self, x, y):
        """ Returns the node that is being dragged/dropped on.

        Returns None if the cursor is not over the icon or label of a node.

        """

        data, wxid, flags, point = self._hit_test((x, y))
        if data is not None:
            populated, node = data

        else:
            node = None

        return node

    def _get_text(self, node):
        """ Returns the tree item text for a node. """

        text = self.model.get_text(node)
        if text is None:
            text = ""

        return text

    def _unpack_event(self, event, wxid=None):
        """ Unpacks the event to see whether a tree item was involved. """

        try:
            point = event.GetPosition()

        except:
            point = event.GetPoint()

        return self._hit_test(point, wxid)

    def _hit_test(self, point, wxid=None):
        """ Determines whether a point is within a node's label or icon. """

        flags = wx.TREE_HITTEST_ONITEMLABEL
        if (wxid is None) or (not wxid.IsOk()):
            wxid, flags = self.control.HitTest(point)

        # Warning: On GTK we have to check the flags before we call 'GetItemData'
        # because if we call it when the hit test returns 'nowhere' it will
        # barf (on Windows it simply returns 'None' 8^()
        if flags & wx.TREE_HITTEST_NOWHERE:
            data = None

        elif (
            flags & wx.TREE_HITTEST_ONITEMICON
            or flags & wx.TREE_HITTEST_ONITEMLABEL
        ):

            data = self.control.GetItemData(wxid)

        # fixme: Not sure why 'TREE_HITTEST_NOWHERE' doesn't catch everything!
        else:
            data = None

        return data, wxid, flags, point

    def _get_selection(self):
        """ Returns a list of the selected nodes """

        selection = []
        for wxid in self.control.GetSelections():
            data = self.control.GetItemData(wxid)
            if data is not None:
                populated, node = data
                selection.append(self.model.get_selection_value(node))

        return selection

    def _expand_item(self, wxid):
        """ Recursively expand a tree item. """

        self.control.Expand(wxid)

        cid, cookie = self.control.GetFirstChild(wxid)
        while cid.IsOk():
            self._expand_item(cid)
            cid, cookie = self.control.GetNextChild(wxid, cookie)

        return

    # Trait event handlers -------------------------------------------------

    def _on_root_changed(self, root):
        """ Called when the root of the model has changed. """

        # Delete everything...
        if self.control is not None:
            self.control.DeleteAllItems()

            self._node_to_id_map = {}

            # ... and then add the root item back in.
            if root is not None:
                self._add_root_node(root)

    def _on_nodes_changed(self, event):
        """ Called when nodes have been changed. """

        self._update_node(self._get_wxid(event.node), event.node)

        for child in event.children:
            cid = self._get_wxid(child)
            if cid is not None:
                self._update_node(cid, child)

    def _on_nodes_inserted(self, event):
        """ Called when nodes have been inserted. """

        parent = event.node
        children = event.children
        index = event.index

        # Has the node actually appeared in the tree yet?
        pid = self._get_wxid(parent)
        if pid is not None:
            # The item data is a tuple.  The first element indicates whether or
            # not we have already populated the item with its children.  The
            # second element is the actual item data.
            if self.show_root or parent is not self.root:
                populated, node = self.control.GetItemData(pid)

            else:
                populated = True

            # If the node is not yet populated then just get the children and
            # add them.
            if not populated:
                for child in self._get_children(parent):
                    self._add_node(pid, child)

            # Otherwise, insert them.
            else:
                # An index of -1 means append!
                if index == -1:
                    index = self.control.GetChildrenCount(pid, False)

                for child in children:
                    self._insert_node(pid, child, index)
                    index += 1

            # The element is now populated!
            if self.show_root or parent is not self.root:
                self.control.SetItemData(pid, (True, parent))

            # Does the node have any children now?
            has_children = self.control.GetChildrenCount(pid) > 0
            self.control.SetItemHasChildren(pid, has_children)

            # If the node is not expanded then expand it.
            if not self.is_expanded(parent):
                self.expand(parent)

    def _on_nodes_removed(self, event):
        """ Called when nodes have been removed. """

        parent = event.node
        children = event.children

        # Has the node actually appeared in the tree yet?
        pid = self._get_wxid(parent)
        if pid is not None:
            for child in event.children:
                cid = self._get_wxid(child)
                if cid is not None:
                    self.control.Delete(cid)

            # Does the node have any children left?
            has_children = self.control.GetChildrenCount(pid) > 0
            self.control.SetItemHasChildren(pid, has_children)

    def _on_nodes_replaced(self, event):
        """ Called when nodes have been replaced. """

        for old_child, new_child in zip(event.old_children, event.children):
            cid = self._get_wxid(old_child)
            if cid is not None:
                # Remove listeners from the old node.
                self.model.remove_listener(old_child)

                # Delete all of the node's children.
                self.control.DeleteChildren(cid)

                # Update the visual appearance of the node.
                self._update_node(cid, new_child)

                # Update the node data.
                #
                # The item data is a tuple.  The first element indicates
                # whether or not we have already populated the item with its
                # children. The second element is the actual item data (which
                # in our case is an arbitrary Python object provided by the
                # tree model).
                self.control.SetItemData(cid, (False, new_child))

                # Remove the old node from the node to Id map.
                self._remove_wxid(old_child)

                # Add the new node to the node to Id map.
                self._set_wxid(new_child, cid)

                # Add listeners to the new node.
                self.model.add_listener(new_child)

                # Does the new node have any children?
                has_children = self._has_children(new_child)
                self.control.SetItemHasChildren(cid, has_children)

        # Update the tree's selection (in case the old node that was replaced
        # was selected, the selection should now include the new node).
        self.selection = self._get_selection()

    def _on_structure_changed(self, event):
        """ Called when the structure of a node has changed drastically. """

        self.refresh(event.node)

        return

    # wx event handlers ----------------------------------------------------

    def _on_char(self, event):
        """ Called when a key is pressed when the tree has focus. """

        self.key_pressed = KeyPressedEvent(
            alt_down=event.altDown,
            control_down=event.controlDown,
            shift_down=event.shiftDown,
            key_code=event.KeyCode,
        )

        event.Skip()

    def _on_left_down(self, event):
        """ Called when the left mouse button is clicked on the tree. """

        data, id, flags, point = self._unpack_event(event)

        # Save point for tree_begin_drag method to workaround a bug in ?? when
        # wx.TreeEvent.GetPoint returns only (0,0).  This happens under linux
        # when using wx-2.4.2.4, for instance.
        self._point_left_clicked = point

        # Did the left click occur on a tree item?
        if data is not None:
            populated, node = data

            # Trait event notification.
            self.node_left_clicked = node, point

        # Give other event handlers a chance.
        event.Skip()

    def _on_right_down(self, event):
        """ Called when the right mouse button is clicked on the tree. """

        data, id, flags, point = self._unpack_event(event)

        # Did the right click occur on a tree item?
        if data is not None:
            populated, node = data

            # Trait event notification.
            self.node_right_clicked = node, point

        # Otherwise notify that the control itself was clicked
        else:
            self.control_right_clicked = point

        # Give other event handlers a chance.
        event.Skip()

    def _on_tree_item_activated(self, event):
        """ Called when a tree item is activated (i.e., double clicked). """

        # fixme: See the comment where the events are wired up for more
        # information.

        ##         # Which item was activated?
        ##         wxid = event.GetItem()

        # Which item was activated.
        point = event.GetPosition()
        wxid, flags = self.control.HitTest(point)

        # The item data is a tuple.  The first element indicates whether or not
        # we have already populated the item with its children.  The second
        # element is the actual item data.
        populated, node = self.control.GetItemData(wxid)

        # Trait event notiification.
        self.node_activated = node

    def _on_tree_item_collapsing(self, event):
        """ Called when a tree item is about to collapse. """

        # Which item is collapsing?
        wxid = event.GetItem()

        # The item data is a tuple.  The first element indicates whether or not
        # we have already populated the item with its children.  The second
        # element is the actual item data.
        populated, node = self.control.GetItemData(wxid)

        # Give the model a chance to veto the collapse.
        if not self.model.is_collapsible(node):
            event.Veto()

    def _on_tree_item_collapsed(self, event):
        """ Called when a tree item has been collapsed. """

        # Which item was collapsed?
        wxid = event.GetItem()

        # The item data is a tuple.  The first element indicates whether or not
        # we have already populated the item with its children.  The second
        # element is the actual item data.
        populated, node = self.control.GetItemData(wxid)

        # Make sure that the item's 'closed' icon is displayed etc.
        self._update_node(wxid, node)

        # Trait event notification.
        self.node_collapsed = node

    def _on_tree_item_expanding(self, event):
        """ Called when a tree item is about to expand. """

        # Which item is expanding?
        wxid = event.GetItem()

        # The item data is a tuple.  The first element indicates whether or not
        # we have already populated the item with its children.  The second
        # element is the actual item data.
        populated, node = self.control.GetItemData(wxid)

        # Give the model a chance to veto the expansion.
        if self.model.is_expandable(node):
            # Lazily populate the item's children.
            if not populated:
                # Add the child nodes.
                for child in self._get_children(node):
                    self._add_node(wxid, child)

                # The element is now populated!
                self.control.SetItemData(wxid, (True, node))

        else:
            event.Veto()

    def _on_tree_item_expanded(self, event):
        """ Called when a tree item has been expanded. """

        # Which item was expanded?
        wxid = event.GetItem()

        # The item data is a tuple.  The first element indicates whether or not
        # we have already populated the item with its children.  The second
        # element is the actual item data.
        populated, node = self.control.GetItemData(wxid)

        # Make sure that the node's 'open' icon is displayed etc.
        self._update_node(wxid, node)

        # Trait event notification.
        self.node_expanded = node

    def _on_tree_begin_label_edit(self, event):
        """ Called when the user has started editing an item's label. """

        wxid = event.GetItem()

        # The item data is a tuple.  The first element indicates whether or not
        # we have already populated the item with its children.  The second
        # element is the actual item data.
        populated, node = self.control.GetItemData(wxid)

        # Give the model a chance to veto the edit.
        if not self.model.is_editable(node):
            event.Veto()

    def _on_tree_end_label_edit(self, event):
        """ Called when the user has finished editing am item's label. """

        wxid = event.GetItem()

        # The item data is a tuple.  The first element indicates whether or not
        # we have already populated the item with its children.  The second
        # element is the actual item data.
        populated, node = self.control.GetItemData(wxid)

        # Give the model a chance to veto the edit.
        label = event.GetLabel()

        # Making sure the new label is not an empty string

        if (
            label is not None
            and len(label) > 0
            and self.model.can_set_text(node, label)
        ):

            def end_label_edit():
                """ Called to complete the label edit. """

                # Set the node's text.
                self.model.set_text(node, label)

                # If a label edit callback was specified (in the call to
                # 'edit_label'), then call it).
                if self._label_edit_callback is not None:
                    self._label_edit_callback(self, node, label)

                return

            # We use a deffered call here, because a name change can trigger
            # the structure of a node to change, and hence the actual tree
            # nodes might get moved/deleted before the label edit operation has
            # completed.  When this happens wx gets very confused!  By using
            # 'invoke_later' we allow the label edit to complete.
            GUI.invoke_later(end_label_edit)

        else:
            event.Veto()

            # If a label edit callback was specified (in the call to
            # 'edit_label'), then call it).
            if self._label_edit_callback is not None:
                self._label_edit_callback(self, node, label)

    def _on_tree_begin_drag(self, event):
        """ Called when a drag operation is starting on a tree item. """

        # Get the node, its id and the point where the event occurred.
        data, wxid, flags, point = self._unpack_event(event, event.GetItem())

        if point == (0, 0):
            # Apply workaround for GTK.
            point = self.point_left_clicked
            wxid, flags = self.HitTest(point)
            data = self.control.GetItemData(wxid)

        if data is not None:
            populated, node = data

            # Give the model a chance to veto the drag.
            if self.model.is_draggable(node):
                # We ask the model for the actual value to drag.
                drag_value = self.model.get_drag_value(node)

                # fixme: This is a terrible hack to get the binding x passed
                # during a drag operation.  Bindings should probably *always*
                # be dragged and our drag and drop mechanism should allow
                # extendable ways to extract the actual data.
                from pyface.wx.drag_and_drop import clipboard

                clipboard.node = [node]

                # Make sure that the tree selection is updated before we start
                # the drag. If we don't do this then if the first thing a
                # user does is drag a tree item (i.e., without a separate click
                # to select it first) then the selection appears empty.
                self.selection = self._get_selection()

                # Start the drag.
                PythonDropSource(self.control, drag_value, self)

                # Trait event notification.
                self.node_begin_drag = node

            else:
                event.Veto()

        return

    # fixme: This is part of the drag and drop hack...
    def on_dropped(self):
        """ Callback invoked when a drag/drop operation has completed. """

        from pyface.wx.drag_and_drop import clipboard

        clipboard.node = None

    def _on_tree_sel_changed(self, event):
        """ Called when the selection is changed. """

        # Update our record of the selection to whatever was selected in the
        # tree UNLESS we are ignoring selection events.
        if not self._ignore_selection_events:

            # Trait notification.
            self.selection = self._get_selection()

    def _on_tree_delete_item(self, event):
        """ Called when a tree item is being been deleted. """

        # Which item is being deleted?
        wxid = event.GetItem()

        # Check if GetItemData() returned a valid to tuple to unpack
        # ...if so, remove the node from the tree, otherwise just return
        #
        # fixme: Whoever addeed this code (and the comment above) didn't say
        # when this was occurring. This is method is called in response to a wx
        # event to delete an item and hence the item data should never be None
        # surely?!? Was it happening just on one platform?!?
        if self.control is None:
            return
        try:
            data = self.control.GetItemData(wxid)
        except Exception:
            # most likely control is in the process of being destroyed
            data = None

        if data is not None:
            # The item data is a tuple.  The first element indicates whether or
            # not we have already populated the item with its children.  The
            # second element is the actual item data.
            populated, node = data

            # Remove the node.
            self._remove_node(wxid, node)



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
import unittest

from traits.etsconfig.api import ETSConfig

from pyface.base_toolkit import find_toolkit, import_toolkit


class TestToolkit(unittest.TestCase):
    def test_import_null_toolkit(self):
        toolkit = import_toolkit("null")
        self.assertEqual(toolkit.package, "pyface")
        self.assertEqual(toolkit.toolkit, "null")

    def test_missing_toolkit(self):
        # test that we get an error with an undefined toolkit
        with self.assertRaises(RuntimeError):
            import_toolkit("nosuchtoolkit")

    def test_find_current_toolkit_no_etsconfig(self):
        old_etsconfig_toolkit = ETSConfig._toolkit
        ETSConfig._toolkit = ""
        try:
            toolkit = find_toolkit("pyface.toolkits", old_etsconfig_toolkit)
            self.assertEqual(toolkit.package, "pyface")
            self.assertEqual(toolkit.toolkit, old_etsconfig_toolkit)
            self.assertEqual(ETSConfig.toolkit, old_etsconfig_toolkit)
        finally:
            ETSConfig._toolkit = old_etsconfig_toolkit

    def test_find_null_toolkit_no_etsconfig(self):
        old_etsconfig_toolkit = ETSConfig._toolkit
        ETSConfig._toolkit = ""
        try:
            toolkit = find_toolkit("pyface.toolkits", "null")
            self.assertEqual(toolkit.package, "pyface")
            self.assertEqual(toolkit.toolkit, "null")
            self.assertEqual(ETSConfig.toolkit, "null")
        finally:
            ETSConfig._toolkit = old_etsconfig_toolkit

    def test_find_nonexistent_toolkit_no_etsconfig(self):
        old_etsconfig_toolkit = ETSConfig._toolkit
        ETSConfig._toolkit = ""
        try:
            toolkit = find_toolkit("pyface.toolkits", "nonexistent")
            self.assertEqual(toolkit.package, "pyface")
            self.assertEqual(toolkit.toolkit, "null")
            self.assertEqual(ETSConfig.toolkit, "null")
        finally:
            ETSConfig._toolkit = old_etsconfig_toolkit



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!


from pyface.action.api import Action, ActionItem, Group
from traits.api import Any, Instance, List, Property, Str, on_trait_change


class TaskWindowToggleAction(Action):
    """ An action for activating an application window.
    """

    # 'Action' interface -----------------------------------------------------

    #: The name of the action for the window.
    name = Property(Str, depends_on="window.title")

    #: The action is a toggle action.
    style = "toggle"

    # 'TaskWindowToggleAction' interface -------------------------------------

    # The window to use for this action.
    window = Instance("pyface.tasks.task_window.TaskWindow")

    # -------------------------------------------------------------------------
    # 'Action' interface.
    # -------------------------------------------------------------------------

    def perform(self, event=None):
        if self.window:
            self.window.activate()

    # -------------------------------------------------------------------------
    # Private interface.
    # -------------------------------------------------------------------------

    def _get_name(self):
        if self.window.title:
            return self.window.title
        return u""

    @on_trait_change("window:activated")
    def _window_activated(self):
        self.checked = True

    @on_trait_change("window:deactivated")
    def _window_deactivated(self):
        self.checked = False


class TaskWindowToggleGroup(Group):
    """ A Group for toggling the activation state of an application's windows.
    """

    # 'Group' interface ------------------------------------------------------

    #: The id of the action group.
    id = "TaskWindowToggleGroup"

    #: The actions in the action group
    items = List()

    # 'TaskWindowToggleGroup' interface --------------------------------------

    #: The application that contains the group.
    application = Instance("pyface.tasks.tasks_application.TasksApplication")

    #: The ActionManager to which the group belongs.
    manager = Any()

    # -------------------------------------------------------------------------
    # 'Group' interface.
    # -------------------------------------------------------------------------

    def destroy(self):
        """ Called when the group is no longer required.
        """
        super(TaskWindowToggleGroup, self).destroy()
        if self.application:
            self.application.on_trait_change(
                self._rebuild, "windows[]", remove=True
            )

    # -------------------------------------------------------------------------
    # Private interface.
    # -------------------------------------------------------------------------

    def _get_items(self):
        items = []
        for window in self.application.windows:
            active = window == self.application.active_window
            action = TaskWindowToggleAction(window=window, checked=active)
            items.append(ActionItem(action=action))
        return items

    def _rebuild(self):
        # Clear out the old group, then build the new one.
        for item in self.items:
            item.destroy()
        self.items = self._get_items()

        # Inform our manager that it needs to be rebuilt.
        self.manager.changed = True

    # Trait initializers -----------------------------------------------------

    def _items_default(self):
        self.application.on_trait_change(self._rebuild, "windows[]")
        return self._get_items()

    def _manager_default(self):
        manager = self
        while isinstance(manager, Group):
            manager = manager.parent
        return manager



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
""" An action that resets *all* perspectives. """


from pyface.api import YES


from .workbench_action import WorkbenchAction


# The message used when confirming the action.
MESSAGE = "Do you want to reset ALL perspectives to their defaults?"


class ResetAllPerspectivesAction(WorkbenchAction):
    """ An action that resets *all* perspectives. """

    # 'Action' interface ---------------------------------------------------

    # The action's unique identifier (may be None).
    id = "pyface.workbench.action.reset_all_perspectives"

    # The action's name (displayed on menus/tool bar tools etc).
    name = "Reset All Perspectives"

    # ------------------------------------------------------------------------
    # 'Action' interface.
    # ------------------------------------------------------------------------

    def perform(self, event):
        """ Perform the action. """

        window = self.window

        if window.confirm(MESSAGE) == YES:
            window.reset_all_perspectives()

        return



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!


from .node_event import NodeEvent
from .node_monitor import NodeMonitor
from .node_manager import NodeManager
from .node_tree import NodeTree
from .node_tree_model import NodeTreeModel
from .node_type import NodeType
from .trait_dict_node_type import TraitDictNodeType
from .trait_list_node_type import TraitListNodeType
from .tree_model import TreeModel

# Tree has not yet been ported to qt
from .tree import Tree



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
# (C) Copyright 2007 Riverbank Computing Limited
# This software is provided without warranty under the terms of the BSD license.
# However, when used with the GPL version of PyQt the additional terms described in the PyQt GPL exception also apply


from pyface.qt import QtCore, QtGui


from traits.api import Bool, Enum, Int, provides, Str


from pyface.i_dialog import IDialog, MDialog
from pyface.constant import OK, CANCEL, YES, NO
from .window import Window


# Map PyQt dialog related constants to the pyface equivalents.
_RESULT_MAP = {
    int(QtGui.QDialog.Accepted): OK,
    int(QtGui.QDialog.Rejected): CANCEL,
    int(QtGui.QMessageBox.Ok): OK,
    int(QtGui.QMessageBox.Cancel): CANCEL,
    int(QtGui.QMessageBox.Yes): YES,
    int(QtGui.QMessageBox.No): NO,
}


@provides(IDialog)
class Dialog(MDialog, Window):
    """ The toolkit specific implementation of a Dialog.  See the IDialog
    interface for the API documentation.
    """

    # 'IDialog' interface -------------------------------------------------#

    cancel_label = Str()

    help_id = Str()

    help_label = Str()

    ok_label = Str()

    resizeable = Bool(True)

    return_code = Int(OK)

    style = Enum("modal", "nonmodal")

    # 'IWindow' interface -------------------------------------------------#

    title = Str("Dialog")

    # ------------------------------------------------------------------------
    # Protected 'IDialog' interface.
    # ------------------------------------------------------------------------

    def _create_buttons(self, parent):
        buttons = QtGui.QDialogButtonBox()

        # 'OK' button.
        if self.ok_label:
            btn = buttons.addButton(
                self.ok_label, QtGui.QDialogButtonBox.AcceptRole
            )
        else:
            btn = buttons.addButton(QtGui.QDialogButtonBox.Ok)

        btn.setDefault(True)
        btn.clicked.connect(self.control.accept)

        # 'Cancel' button.
        if self.cancel_label:
            btn = buttons.addButton(
                self.cancel_label, QtGui.QDialogButtonBox.RejectRole
            )
        else:
            btn = buttons.addButton(QtGui.QDialogButtonBox.Cancel)

        btn.clicked.connect(self.control.reject)

        # 'Help' button.
        # FIXME v3: In the original code the only possible hook into the help
        # was to reimplement self._on_help().  However this was a private
        # method.  Obviously nobody uses the Help button.  For the moment we
        # display it but can't actually use it.
        if len(self.help_id) > 0:
            if self.help_label:
                buttons.addButton(
                    self.help_label, QtGui.QDialogButtonBox.HelpRole
                )
            else:
                buttons.addButton(QtGui.QDialogButtonBox.Help)

        return buttons

    def _create_contents(self, parent):
        layout = QtGui.QVBoxLayout()

        if not self.resizeable:
            layout.setSizeConstraint(QtGui.QLayout.SetFixedSize)

        layout.addWidget(self._create_dialog_area(parent))
        layout.addWidget(self._create_buttons(parent))

        parent.setLayout(layout)

    def _create_dialog_area(self, parent):
        panel = QtGui.QWidget(parent)
        panel.setMinimumSize(QtCore.QSize(100, 200))

        palette = panel.palette()
        palette.setColor(QtGui.QPalette.Window, QtGui.QColor("red"))
        panel.setPalette(palette)
        panel.setAutoFillBackground(True)

        return panel

    def _show_modal(self):
        dialog = self.control
        dialog.setWindowModality(QtCore.Qt.ApplicationModal)

        # Suppress the context-help button hint, which
        # results in a non-functional "?" button on Windows.
        dialog.setWindowFlags(
            dialog.windowFlags() & ~QtCore.Qt.WindowContextHelpButtonHint
        )

        retval = dialog.exec_()
        return _RESULT_MAP[retval]

    # ------------------------------------------------------------------------
    # Protected 'IWidget' interface.
    # ------------------------------------------------------------------------

    def _create_control(self, parent):
        dlg = QtGui.QDialog(parent)

        # Setting return code and firing close events is handled for 'modal' in
        # MDialog's open method. For 'nonmodal', we do it here.
        if self.style == "nonmodal":
            dlg.finished.connect(self._finished_fired)

        if self.size != (-1, -1):
            dlg.resize(*self.size)

        if self.position != (-1, -1):
            dlg.move(*self.position)

        dlg.setWindowTitle(self.title)

        return dlg

    # ------------------------------------------------------------------------
    # Private interface.
    # ------------------------------------------------------------------------

    def _finished_fired(self, result):
        """ Called when the dialog is closed (and nonmodal). """

        self.return_code = _RESULT_MAP[result]
        self.close()



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
from .bracket_matcher import BracketMatcher
from .call_tip_widget import CallTipWidget
from .completion_lexer import CompletionLexer
from .console_widget import ConsoleWidget
from .history_console_widget import HistoryConsoleWidget



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
from . import qt_api

if qt_api == "pyqt":
    from PyQt4.QtCore import *

    from PyQt4.QtCore import pyqtProperty as Property
    from PyQt4.QtCore import pyqtSignal as Signal
    from PyQt4.QtCore import pyqtSlot as Slot
    from PyQt4.Qt import QCoreApplication
    from PyQt4.Qt import Qt

    __version__ = QT_VERSION_STR
    __version_info__ = tuple(map(int, QT_VERSION_STR.split(".")))

elif qt_api == "pyqt5":
    from PyQt5.QtCore import *

    from PyQt5.QtCore import pyqtProperty as Property
    from PyQt5.QtCore import pyqtSignal as Signal
    from PyQt5.QtCore import pyqtSlot as Slot
    from PyQt5.Qt import QCoreApplication
    from PyQt5.Qt import Qt

    __version__ = QT_VERSION_STR
    __version_info__ = tuple(map(int, QT_VERSION_STR.split(".")))


elif qt_api == "pyside2":
    from PySide2.QtCore import *

    from PySide2 import __version__, __version_info__
else:
    try:
        from PySide import __version__, __version_info__
    except ImportError:
        pass
    from PySide.QtCore import *



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest
from unittest import mock

from pyface.tasks.api import TaskLayout, PaneItem
from pyface.toolkit import toolkit_object
from pyface.window import Window

try:
    from pyface.qt import QtGui
    from pyface.ui.qt4.tasks.main_window_layout import MainWindowLayout
except ImportError:
    if toolkit_object.toolkit == "qt4":
        raise


GuiTestAssistant = toolkit_object("util.gui_test_assistant:GuiTestAssistant")


def create_dummy_dock_widget(parent):
    """ Create a dummy QDockWidget with a dummy child widget for test.

    Parameters
    ----------
    parent : QObject

    Returns
    -------
    dock_widget : QDockWidget
    """
    dock_widget = QtGui.QDockWidget(parent)
    content_widget = QtGui.QWidget(parent)
    dock_widget.setWidget(content_widget)
    return dock_widget


@unittest.skipIf(
    toolkit_object.toolkit != "qt4",
    "This test targets Qt specific MainWindowLayout. "
    "Current toolkit is not Qt."
)
class TestMainWindowLayout(unittest.TestCase, GuiTestAssistant):
    """ Test Qt specific MainWindowLayout.

    Note that MainWindowLayout does not have a toolkit-agnostic interface
    in the ``pyface.tasks`` package. Therefore this test is Qt-only.
    """

    def setUp(self):
        GuiTestAssistant.setUp(self)
        self.window = Window(size=(500, 500))
        self.window._create()

    def tearDown(self):
        if self.window.control is not None:
            with self.delete_widget(self.window.control):
                self.window.destroy()
        del self.window
        GuiTestAssistant.tearDown(self)

    def setup_window_with_central_widget(self):
        # Add a central widget to the main window.
        # The main window takes ownership of the child widget.
        central_widget = QtGui.QWidget(parent=self.window.control)
        self.window.control.setCentralWidget(central_widget)

    def test_set_pane_item_width_in_main_window_layout(self):
        # Test the dock pane width is as expected.

        self.setup_window_with_central_widget()

        # Set the dock widget expected width to be smaller than the window
        # for a meaningful test.
        expected_width = self.window.size[0] // 2
        window_layout = MainWindowLayout(control=self.window.control)
        dock_layout = TaskLayout(
            left=PaneItem(width=expected_width)
        )
        dock_widget = create_dummy_dock_widget(parent=self.window.control)
        patch_get_dock_widget = mock.patch.object(
            MainWindowLayout, "_get_dock_widget",
            return_value=dock_widget,
        )

        # when
        with self.event_loop():
            with patch_get_dock_widget:
                window_layout.set_layout(dock_layout)

        # then
        size = dock_widget.widget().size()
        self.assertEqual(size.width(), expected_width)

    def test_set_pane_item_height_in_main_window_layout(self):
        # Test the dock pane height is as expected.

        self.setup_window_with_central_widget()

        # Set the dock widget expected height to be smaller than the window
        # for a meaningful test.
        expected_height = self.window.size[1] // 2
        window_layout = MainWindowLayout(control=self.window.control)
        dock_layout = TaskLayout(
            bottom=PaneItem(height=expected_height)
        )
        dock_widget = create_dummy_dock_widget(parent=self.window.control)
        patch_get_dock_widget = mock.patch.object(
            MainWindowLayout, "_get_dock_widget",
            return_value=dock_widget,
        )

        # when
        with self.event_loop():
            with patch_get_dock_widget:
                window_layout.set_layout(dock_layout)

        # then
        size = dock_widget.widget().size()
        self.assertEqual(size.height(), expected_height)



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" The implementation of a dialog that prompts the user for confirmation. """


from .constant import NO


def confirm(parent, message, title=None, cancel=False, default=NO):
    """ Convenience method to show a confirmation dialog.

    Parameters
    ----------
    parent : toolkit widget or None
        The parent control for the dialog.
    message : str
        The text of the message to display.
    title : str
        The text of the dialog title.
    cancel : bool
        ``True`` if the dialog should contain a Cancel button.
    default : NO, YES or CANCEL
        Which button should be the default button.
    """
    if title is None:
        title = "Confirmation"

    dialog = ConfirmationDialog(
        parent=parent,
        message=message,
        cancel=cancel,
        default=default,
        title=title,
    )

    return dialog.open()


# Import the toolkit specific version.
from .toolkit import toolkit_object

ConfirmationDialog = toolkit_object("confirmation_dialog:ConfirmationDialog")



# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from traits.api import (
    Any,
    DelegatesTo,
    HasTraits,
    Instance,
    Interface,
    provides,
)


class ITaskWindowBackend(Interface):
    """ The TaskWindow layout interface.

    TaskWindow delegates to an ITaskWindowBackend object for toolkit-specific
    layout functionality.
    """

    #: The root control of the TaskWindow to which the layout belongs.
    control = Any()

    #: The TaskWindow to which the layout belongs.
    window = Instance("pyface.tasks.task_window.TaskWindow")

    # ------------------------------------------------------------------------
    # 'ITaskWindowBackend' interface.
    # ------------------------------------------------------------------------

    def create_contents(self, parent):
        """ Create and return the TaskWindow's contents. (See IWindow.)
        """

    def destroy(self):
        """ Destroy the backend.

        Note that TaskWindow will destroy the widget created in create_contents,
        but this method may be used to perform additional cleanup.
        """

    def hide_task(self, state):
        """ Assuming the specified TaskState is active, hide its controls.
        """

    def show_task(self, state):
        """ Assuming no task is currently active, show the controls of the
            specified TaskState.
        """

    # Methods for saving and restoring the layout -------------------------#

    def get_layout(self):
        """ Returns a TaskLayout for the current state of the window.
        """

    def set_layout(self, layout):
        """ Applies a TaskLayout (which should be suitable for the active task)
            to the window.
        """


@provides(ITaskWindowBackend)
class MTaskWindowBackend(HasTraits):
    """ Mixin containing common coe for toolkit-specific implementations.
    """

    # 'ITaskWindowBackend' interface ---------------------------------------

    control = DelegatesTo("window")
    window = Instance("pyface.tasks.task_window.TaskWindow")

    # ------------------------------------------------------------------------
    # 'ITaskWindowBackend' interface.
    # ------------------------------------------------------------------------

    def create_contents(self, parent):
        raise NotImplementedError

    def destroy(self):
        pass

    def hide_task(self, state):
        raise NotImplementedError

    def show_task(self, state):
        raise NotImplementedError

    def get_layout(self):
        raise NotImplementedError

    def set_layout(self, layout):
        raise NotImplementedError
